@@ -1298,13 +1298,13 @@ g_usb_device_finalize (GObject *object)
 GUsbDevice *
 _g_usb_device_new (GUsbContext *context,
 		   libusb_device *device,
-		   GUdevDevice *udev)
+		   const gchar *platform_id)
 {
 	GObject *obj;
 	obj = g_object_new (G_USB_TYPE_DEVICE,
 			    "context", context,
 			    "libusb-device", device,
-			    "platform-id", g_udev_device_get_sysfs_path (udev),
+			    "platform-id", platform_id,
 			    NULL);
 	return G_USB_DEVICE (obj);
 }
@@ -1420,6 +1420,22 @@ g_usb_device_get_manufacturer_index (GUsbDevice *device)
 	return device->priv->desc.iManufacturer;
 }
 
+/**
+ * g_usb_device_get_device_class:
+ * @device: a #GUsbDevice
+ *
+ * Gets the device class.
+ *
+ * Return value: a device class number, e.g. 0x09 is a USB hub.
+ *
+ * Since: 0.1.7
+ **/
+guint8
+g_usb_device_get_device_class (GUsbDevice *device)
+{
+	return device->priv->desc.bDeviceClass;
+}
+
 /**
  * g_usb_device_get_product_index:
  * @device: a #GUsbDevice
