@@ -30,7 +30,6 @@
 #include "config.h"
 
 #include <libusb-1.0/libusb.h>
-#include <poll.h>
 #include <stdlib.h>
 
 #include "gusb-context.h"
@@ -39,6 +38,12 @@
 #include "gusb-source.h"
 #include "gusb-source-private.h"
 
+/* the <poll.h> header is not available on all platforms */
+#ifndef POLLIN
+#define POLLIN		0x0001
+#define POLLOUT		0x0004
+#endif
+
 /**
  * g_usb_source_error_quark:
  *
@@ -242,7 +247,7 @@ _g_usb_source_new (GMainContext *main_ctx,
 
 	/* watch the fd's already created */
 	pollfds = libusb_get_pollfds (gusb_source->ctx);
-	for (i=0; pollfds[i] != NULL; i++)
+	for (i = 0; pollfds != NULL && pollfds[i] != NULL; i++)
 		g_usb_source_pollfd_add (gusb_source,
 					 pollfds[i]->fd,
 					 pollfds[i]->events);
